/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.1 Stauverlaufsanalyse
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.anastau.bestimmungstau.indikator;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import com.bitctrl.Constants;

import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.ibv.anastau.bestimmungstau.parameter.ISobParameterListener;
import de.bsvrz.ibv.anastau.bestimmungstau.parameter.ISobParameterObserver;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensatzUpdateEvent;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensatzUpdateListener;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.StoerfallIndikator;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.onlinedaten.OdStoerfallZustand;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.parameter.PdSituationsEigenschaften;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.zustaende.StoerfallSituation;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Die abstrake Implementierung eines Störfallindikators für die
 * Stauobjektbestimmung.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public abstract class SobIndikator implements ISobIndikator, ISobParameterListener, DatensatzUpdateListener {

	/**
	 * Logger für Debugausgaben.
	 */
	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * die letzte Störfall-Situation des Indikators.
	 */
	private SobStoerfallSituation letzteSituation = SobStoerfallSituation.UNDEFINIERT;

	/** die letzte Störfallfallindikator-Situation des Indikators. */
	private StoerfallSituation letzteIndikatorSituation = StoerfallSituation.UNBEKANNT;

	/**
	 * der zu Grunde liegende Störfallindikator innerhalb der
	 * Datenverteilerkonfiguration.
	 */
	private final StoerfallIndikator indikator;

	/**
	 * die Strasse, der der Indikator zugeordnet ist.
	 */
	// private final ISobStrasse strasse;
	/**
	 * die Liste der Verfahren, für die die berechneten Störfallsituationen
	 * ausgewertet werden sollen.
	 */
	private final Set<Aspect> verfahren = new HashSet<>();

	/**
	 * markiert, ob der Indikator bereits für die Berechnung der Stauobjekte
	 * herangezogen wurde.
	 */
	private boolean geaendert;

	/**
	 * die {@link FahrtRichtung} in der der Indikator zur Stauobjektbestimmung
	 * verwendet wird.
	 */
	private final FahrtRichtung richtung;

	/**
	 * die Position des Indikators innerhalb der Fahrtrichtung.
	 */
	private double position;

	/** minimale Güte, mit der Störfallindikatoren berücksichtigt werden. */
	private double minSiGuete;

	/**
	 * Konstruktor.<br>
	 * Die Funktion erzeugt eine Instanz des Indikators.
	 *
	 * @param richtung
	 *            die Fahrtrichtung für die der Indikator zur
	 *            Stauobjektbestimmung verwendet wird.
	 * @param indikator
	 *            der verwendete Störfallindikator aus der
	 *            Datenverteilerkonfiguration
	 */
	public SobIndikator(final FahrtRichtung richtung, final StoerfallIndikator indikator) {
		// this.strasse = strasse;
		this.richtung = richtung;
		this.indikator = indikator;
		final ISobParameterObserver parameter = richtung.getStrasse().getParameter();
		if (parameter != null) {
			parameter.addParameterListener(this);
			anmelden(parameter.getVerfahren());
			minSiGuete = (double) parameter.getMinSiGuete() / Constants.PERCENT;
		}
	}

	/**
	 * meldet den Störfallindikator zum Empfang der Störfallsituationen unter
	 * den gegebenen Aspekten an.
	 *
	 * @param neueVerfahren
	 *            die Liste der Aspekte unter denen Störfallsituationen erfasst
	 *            werden sollen
	 */
	private void anmelden(final Collection<Aspect> neueVerfahren) {
		final Collection<Aspect> entfernteVerfahren = new ArrayList<>();
		for (final Aspect asp : verfahren) {
			if (!neueVerfahren.contains(asp)) {
				indikator.getOnlineDatensatz(OdStoerfallZustand.class).removeUpdateListener(asp, this);
				entfernteVerfahren.add(asp);
			}
		}

		for (final Aspect asp : entfernteVerfahren) {
			verfahren.remove(asp);
		}

		for (final Aspect asp : neueVerfahren) {
			if (!verfahren.contains(asp)) {
				indikator.getOnlineDatensatz(OdStoerfallZustand.class).addUpdateListener(asp, this);
			}
		}
		verfahren.addAll(neueVerfahren);
	}

	@Override
	public final void changed(final ISobParameterObserver parameter) {
		anmelden(parameter.getVerfahren());
		minSiGuete = (double) parameter.getMinSiGuete() / Constants.PERCENT;
	}

	@Override
	public void datensatzAktualisiert(final DatensatzUpdateEvent event) {
		if (event.getDatensatz() instanceof PdSituationsEigenschaften) {
			SobIndikator.LOGGER.finer("Neue Situation für Indikator " + indikator + ": " + event.getDatum());
		}
	}

	/**
	 * liefert den aktuellen Zustand des zu Grunde liegenden Indikators.<br>
	 * Es wird das Maximum aus allen für die übergebenen Aspekte bestehenden
	 * Zustandswerten ermittelt und geliefert.
	 *
	 * @param aspekte
	 *            die Aspekte die in die Bestimmung des Zustandes einbezogen
	 *            werden sollen
	 * @return der aktuelle Störfallzustand des Indikators.
	 */
	private StoerfallSituation getIndikatorSituation(final Set<Aspect> aspekte) {
		StoerfallSituation situation = StoerfallSituation.UNBEKANNT;
		final OdStoerfallZustand datenSatz = indikator.getOnlineDatensatz(OdStoerfallZustand.class);
		if (datenSatz != null) {
			for (final Aspect asp : aspekte) {
				final OdStoerfallZustand.Daten daten = datenSatz.getDatum(asp);
				if (daten != null) {
					final StoerfallSituation s = daten.getSituation();
					if (daten.getGueteIndex() >= minSiGuete) {
						if ((s != null) && (s.getCode() > situation.getCode())) {
							situation = s;
						}
					}
				}
			}
		}

		return situation;
	}

	@Override
	public final String getName() {
		return indikator.getSystemObject().getNameOrPidOrId();
	}

	@Override
	public final ISobIndikator getNext() {
		ISobIndikator result = null;
		final int index = richtung.indexOf(this) + 1;

		if (index < richtung.anzahlIndikatoren()) {
			result = richtung.getIndikator(index);
		}

		return result;
	}

	@Override
	public double getPosition() {
		richtung.berechnePositionen();
		return position;
	}

	@Override
	public final ISobIndikator getPrev() {
		ISobIndikator result = null;
		final int index = richtung.indexOf(this) - 1;

		if (index >= 0) {
			result = richtung.getIndikator(index);
		}

		return result;
	}

	@Override
	public final IFahrtRichtung getRichtung() {
		return richtung;
	}

	@Override
	public final SobStoerfallSituation getSituation() {
		return letzteSituation;
	}

	@Override
	public final ISobStrasse getStrasse() {
		return richtung.getStrasse();
	}

	/**
	 * liefert den verwendeten Störfallindikator.
	 *
	 * @return den Indikator
	 */
	protected final StoerfallIndikator getUsedIndikator() {
		return indikator;
	}

	@Override
	public boolean isFreierVerkehr() {
		return letzteIndikatorSituation == StoerfallSituation.FREIER_VERKEHR;
	}

	@Override
	public final boolean isGeaendert() {
		return geaendert;
	}

	@Override
	public final void setGeaendert(final boolean status) {
		this.geaendert = status;
	}

	@Override
	public final void setPosition(final double position) {
		this.position = position;
	}

	@Override
	public final void sichereSituation() {
		letzteIndikatorSituation = getIndikatorSituation(verfahren);
		SobStoerfallSituation neueSobSituation = SobStoerfallSituation.UNDEFINIERT;
		switch (letzteIndikatorSituation) {
		case STAU:
			neueSobSituation = SobStoerfallSituation.STAU;
			break;
		case UNBEKANNT:
		case STOERUNG:
		case KEINE_AUSSAGE:
			neueSobSituation = SobStoerfallSituation.UNDEFINIERT;
			break;
		default:
			neueSobSituation = SobStoerfallSituation.VERKEHR;
			break;
		}

		if (neueSobSituation != letzteSituation) {
			letzteSituation = neueSobSituation;
			geaendert = true;
		}
	}

	@Override
	public String toString() {
		return indikator.toString();
	}
}
