/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.1 Stauverlaufsanalyse
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.anastau.progstau;

import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import com.bitctrl.Constants;
import com.bitctrl.util.Interval;

import de.bsvrz.iav.gllib.gllib.dav.GanglinieMQ;
import de.bsvrz.iav.gllib.gllib.dav.Ganglinienprognose;
import de.bsvrz.iav.gllib.gllib.dav.GlProgAnfrage;
import de.bsvrz.iav.gllib.gllib.dav.GlProgAntwortEvent;
import de.bsvrz.iav.gllib.gllib.dav.GlProgAntwortListener;
import de.bsvrz.iav.gllib.gllib.dav.Messwerte;
import de.bsvrz.ibv.anastau.anastau.StauverlaufsAnalyse;
import de.bsvrz.ibv.anastau.anastau.SvaMessage;
import de.bsvrz.ibv.anastau.bestimmungstau.indikator.SobMessQuerschnitt;
import de.bsvrz.ibv.anastau.bestimmungstau.indikator.SobStoerfallSituation;
import de.bsvrz.ibv.anastau.bestimmungstau.indikator.SobStrassenKnoten;
import de.bsvrz.ibv.anastau.bestimmungstau.stau.IStau;
import de.bsvrz.ibv.anastau.bestimmungstau.stau.Stau;
import de.bsvrz.sys.funclib.bitctrl.daf.BetriebsmeldungDaten;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensendeException;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.sys.funclib.bitctrl.util.UndefiniertException;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Klasse zur Ermittlung der Zuflussmenge zu einem Stau.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class SvpZuflussMenge implements ISvpZuflussMenge, GlProgAntwortListener {

	/**
	 * Logger für Debugausgaben.
	 */
	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * globaler Zähler für die Bildung eindeutiger ID für Ganglinienanfragen.
	 */
	private static long lastCounter;

	/**
	 * liefert einen Bezeichner für die nächstfolgende Anfrage an die
	 * Ganglinienprognose.
	 *
	 * @return den Bezeichner
	 */
	private static String getNextAnfrageZeichen() {
		SvpZuflussMenge.lastCounter++;
		return SvpZuflussMenge.class.getName() + "_" + SvpZuflussMenge.lastCounter;
	}

	/**
	 * die Ganglinie, die für die Ermittlung der Zuflussmenge genutzt werden
	 * soll.
	 */
	private GanglinieMQ gangLinie;

	/**
	 * der Messquerschnitt, der die Daten für die Ermittlung der Zuflussmenge
	 * liefern soll.
	 */
	private MessQuerschnittAllgemein mq;

	/**
	 * die Messquerschnitte auf Zufahrten vor dem Stau, die Daten für die
	 * Ermittlung der Zuflussmenge liefern sollen.
	 */
	private final Map<MessQuerschnittAllgemein, GanglinieMQ> zufahrtenDavor = new HashMap<>();

	/**
	 * die Messquerschnitte auf Abfahrten vor dem Stau, die Daten für die
	 * Ermittlung der Zuflussmenge liefern sollen.
	 */
	private final Map<MessQuerschnittAllgemein, GanglinieMQ> abfahrtenDavor = new HashMap<>();

	/**
	 * die Messquerschnitte auf Zufahrten im Stau, die Daten für die Ermittlung
	 * der Zuflussmenge liefern sollen.
	 */
	private final Map<MessQuerschnittAllgemein, GanglinieMQ> zufahrtenImStau = new HashMap<>();

	/**
	 * die aktuellen QB-Werte für Messquerschnitte auf Zu- und Abfahrten im
	 * Stau.
	 */
	private final Map<MessQuerschnittAllgemein, Number> qbWerteMq = new HashMap<>();

	/**
	 * die Messquerschnitte auf Abfahrten im Stau, die Daten für die Ermittlung
	 * der Zuflussmenge liefern sollen.
	 */
	private final Map<MessQuerschnittAllgemein, GanglinieMQ> abfahrtenImStau = new HashMap<>();

	/**
	 * der Stau für den der Zufluss ermittelt werden soll.
	 */
	private final IStau stau;

	/**
	 * die verwendete Ganglinienprognose.
	 */
	private Ganglinienprognose prognose;

	/**
	 * das Kennzeichen für die verwendete Anfrage.
	 */
	private String anfrageZeichen;

	/** die Startzeit für die die Zuflussmengen berechnet werden sollen. */
	private final long startZeit;

	/**
	 * das Intervall das für die lineare Dämpfung der Ganglinien für Ab- und
	 * Zufahrten im Stau verwendet wird. Zuflussdaten benötigt werden.
	 */
	private final long daempfungsZeit;

	/**
	 * Konstruktor.<br>
	 * Die Funktion erzeugt eine Instanz der Klasse zur Zuflussermittlung.
	 *
	 * @param stau
	 *            der Stau, für den dr Zufluss ermittelt werden soll
	 * @param intervall
	 *            das Intervall für das eine Prognose erfolgt und für das
	 *            Zuflussdaten benötigt werden.
	 * @param daempfungsZeit
	 *            das Intervall das für die lineare Dämpfung der Ganglinien für
	 *            Ab- und Zufahrten im Stau verwendet wird. Zuflussdaten
	 *            benötigt werden.
	 */
	public SvpZuflussMenge(final IStau stau, final Interval intervall, final long daempfungsZeit) {
		this.stau = stau;
		this.startZeit = intervall.getStart();
		this.daempfungsZeit = daempfungsZeit;
		final SobMessQuerschnitt sobMq = ermittleMessquerschnitt();

		final Collection<GlProgAnfrage> anfragen = new ArrayList<>();
		if (sobMq != null) {
			mq = sobMq.getIndikator();
			prognose = Ganglinienprognose.getInstanz();
			prognose.addAntwortListener(this);
			anfragen.add(new GlProgAnfrage(mq, intervall, isMqImStau(sobMq)));

			if (StauverlaufsAnalyse.getOptionen().benutzeAnschluesseVorDemStau()) {
				for (final SobStrassenKnoten knoten : stau.ermittleAsAmEnde(mq)) {
					for (final MessQuerschnittAllgemein mqObj : knoten.getAbfahrtMq()) {
						abfahrtenDavor.put(mqObj, null);
						anfragen.add(new GlProgAnfrage(mqObj, intervall, true));
					}
					for (final MessQuerschnittAllgemein mqObj : knoten.getZufahrtMq()) {
						zufahrtenDavor.put(mqObj, null);
						anfragen.add(new GlProgAnfrage(mqObj, intervall, true));
					}
				}
			}

			if (StauverlaufsAnalyse.getOptionen().benutzeAnschluesseImStau()) {
				for (final SobStrassenKnoten knoten : stau.getKnotenImStau()) {
					for (final MessQuerschnittAllgemein mqObj : knoten.getAbfahrtMq()) {
						abfahrtenImStau.put(mqObj, null);
						anfragen.add(new GlProgAnfrage(mqObj, intervall, true));
						qbWerteMq.put(mqObj, StauverlaufsPrognose.getMqQbWert(mqObj));
					}
					for (final MessQuerschnittAllgemein mqObj : knoten.getZufahrtMq()) {
						zufahrtenImStau.put(mqObj, null);
						anfragen.add(new GlProgAnfrage(mqObj, intervall, true));
						qbWerteMq.put(mqObj, StauverlaufsPrognose.getMqQbWert(mqObj));
					}
				}
			}

			if (anfragen.size() > 0) {
				anfrageZeichen = SvpZuflussMenge.getNextAnfrageZeichen();
				synchronized (prognose) {
					try {
						prognose.sendeAnfrage(anfrageZeichen, anfragen);
						prognose.wait(Constants.MILLIS_PER_MINUTE * 2);
					} catch (final InterruptedException e) {
						LogTools.log(SvpZuflussMenge.LOGGER, new BetriebsmeldungDaten(mq.getSystemObject()),
								SvaMessage.WarningGanglinienAbfrage, stau, intervall, e.getMessage());
					} catch (final DatensendeException e) {
						LogTools.log(SvpZuflussMenge.LOGGER, new BetriebsmeldungDaten(mq.getSystemObject()),
								SvaMessage.WarningGanglinienAbfrage, stau, intervall, e.getMessage());
					}

					if (gangLinie == null) {
						LogTools.log(SvpZuflussMenge.LOGGER, new BetriebsmeldungDaten(mq.getSystemObject()),
								SvaMessage.WarningKeineGanglinie, stau, intervall);
						prognose.removeAntwortListener(this);
					}
				}
			}
		}
	}

	/**
	 * Verarbeitung der Antwort auf eine Ganglinienanfrage.
	 *
	 * @param e
	 *            die Antwort von der Ganglinienprognose
	 *
	 * @see de.bsvrz.iav.gllib.gllib.dav.GlProgAntwortListener#antwortEingetroffen(de.bsvrz.iav.gllib.gllib.dav.GlProgAntwortEvent)
	 */
	@Override
	public final void antwortEingetroffen(final GlProgAntwortEvent e) {
		if (anfrageZeichen.equals(e.getAbsenderZeichen())) {
			SvpZuflussMenge.LOGGER.fine("Antwort von Ganglinienprognose eingetroffen: " + e);

			for (final GanglinieMQ gl : e.getGanglinien()) {
				final MessQuerschnittAllgemein glMq = gl.getMessQuerschnitt();
				if (glMq == null) {
					continue;
				}

				if (glMq.equals(mq)) {
					gangLinie = gl;
				}

				for (final MessQuerschnittAllgemein mqObj : new ArrayList<>(
						zufahrtenDavor.keySet())) {
					if (glMq.equals(mqObj)) {
						zufahrtenDavor.put(mqObj, gl);
						break;
					}
				}
				for (final MessQuerschnittAllgemein mqObj : new ArrayList<>(
						abfahrtenDavor.keySet())) {
					if (glMq.equals(mqObj)) {
						abfahrtenDavor.put(mqObj, gl);
						break;
					}
				}
				for (final MessQuerschnittAllgemein mqObj : new ArrayList<>(
						zufahrtenImStau.keySet())) {
					if (glMq.equals(mqObj)) {
						zufahrtenImStau.put(mqObj, gl);
						break;
					}
				}
				for (final MessQuerschnittAllgemein mqObj : new ArrayList<>(
						abfahrtenImStau.keySet())) {
					if (glMq.equals(mqObj)) {
						abfahrtenImStau.put(mqObj, gl);
						break;
					}
				}
			}

			synchronized (prognose) {
				prognose.notifyAll();
			}
			prognose.removeAntwortListener(this);
		}
	}

	/**
	 * ermittelt den Messquerschnitt, der für die Ermittlung des zufluss
	 * verwendet wird.<br>
	 * Es wird auf den ersten Messquerschnitt, der entgegengesetzt zur
	 * Fahrtrichtung zwischen dem Stauende und dem nächsten Autobahnkreuz liegt
	 * zurückgegriffen. Wird auf diese Art keiner gefunden, wird der letzte im
	 * Stau befindliche Messquerschnitt eingesetzt.
	 *
	 * @return den zu verwendenden Messquerschnitt
	 */
	private SobMessQuerschnitt ermittleMessquerschnitt() {
		SobMessQuerschnitt result = null;
		result = stau.getMqDavor();
		if (result == null) {
			result = stau.getMqAmEnde();
		}
		return result;
	}

	@Override
	public final double getZuflussMenge(final long zeitPunkt) throws UndefiniertException {
		if (gangLinie == null) {
			throw new UndefiniertException("Für MQ: \"" + mq + "\" konnte keine Ganglinie ermittelt werden");
		}

		Double wert = null;
		Messwerte messwerte = gangLinie.getStuetzstelle(zeitPunkt).getWert();
		SvpZuflussMenge.LOGGER.finer("Werte für Stützstelle für MQ: " + mq + " um "
				+ DateFormat.getDateTimeInstance().format(new Date(zeitPunkt)) + " = " + messwerte);

		if (messwerte != null) {
			wert = messwerte.getQB();
		}

		if (wert == null) {
			throw new UndefiniertException("Es existiert kein Wert für den Zeitpunkt");
		}

		// Zu- und Abfahrten vor dem Stau berücksichtigen

		for (final Entry<MessQuerschnittAllgemein, GanglinieMQ> eintrag : zufahrtenDavor.entrySet()) {
			if (eintrag.getValue() != null) {
				Double zuflussWert = null;
				messwerte = eintrag.getValue().getStuetzstelle(zeitPunkt).getWert();
				if (messwerte != null) {
					zuflussWert = messwerte.getQB();
				}
				if (zuflussWert == null) {
					throw new UndefiniertException("Es existiert kein Wert für den Zeitpunkt");
				}
				wert = wert + zuflussWert;
			}
		}

		for (final Entry<MessQuerschnittAllgemein, GanglinieMQ> eintrag : abfahrtenDavor.entrySet()) {
			if (eintrag.getValue() != null) {
				Double abflussWert = null;
				messwerte = eintrag.getValue().getStuetzstelle(zeitPunkt).getWert();
				if (messwerte != null) {
					abflussWert = messwerte.getQB();
				}
				if (abflussWert == null) {
					throw new UndefiniertException("Es existiert kein Wert für den Zeitpunkt");
				}
				wert = wert - abflussWert;
			}
		}

		// Zu- und Abfahrten im Stau berücksichtigen
		for (final Entry<MessQuerschnittAllgemein, GanglinieMQ> eintrag : zufahrtenImStau.entrySet()) {
			final Number qbWert = qbWerteMq.get(eintrag.getKey());
			if (qbWert == null) {
				LogTools.log(SvpZuflussMenge.LOGGER, new BetriebsmeldungDaten(eintrag.getKey().getSystemObject()),
						SvaMessage.WarningKeineZufahrtsDatenImStau, eintrag.getKey());
			} else {
				if (eintrag.getValue() != null) {
					Double anfangsWert = null;
					messwerte = eintrag.getValue().getStuetzstelle(startZeit).getWert();
					if (messwerte != null) {
						anfangsWert = messwerte.getQB();
					}
					if (anfangsWert != null) {
						Double zuflussWert = null;
						messwerte = eintrag.getValue().getStuetzstelle(zeitPunkt).getWert();
						if (messwerte != null) {
							zuflussWert = messwerte.getQB();
						}
						if (zuflussWert == null) {
							throw new UndefiniertException("Es existiert kein Wert für den Zeitpunkt");
						}

						final double mqWert = zuflussWert.doubleValue()
								+ ((qbWert.doubleValue() - anfangsWert.doubleValue())
										* (1.0 - ((double) (zeitPunkt - startZeit) / daempfungsZeit)));
						wert = wert + mqWert;
					}
				}
			}
		}

		for (final Entry<MessQuerschnittAllgemein, GanglinieMQ> eintrag : abfahrtenImStau.entrySet()) {
			final Number qbWert = qbWerteMq.get(eintrag.getKey());
			if (qbWert == null) {
				LogTools.log(SvpZuflussMenge.LOGGER, new BetriebsmeldungDaten(eintrag.getKey().getSystemObject()),
						SvaMessage.WarningKeineAbfahrtsDatenImStau, eintrag.getKey());
			} else {
				if (eintrag.getValue() != null) {
					Double anfangsWert = null;
					messwerte = eintrag.getValue().getStuetzstelle(startZeit).getWert();
					if (messwerte != null) {
						anfangsWert = messwerte.getQB();
					}
					if (anfangsWert != null) {
						Double zuflussWert = null;
						messwerte = eintrag.getValue().getStuetzstelle(zeitPunkt).getWert();
						if (messwerte != null) {
							zuflussWert = messwerte.getQB();
						}
						if (zuflussWert == null) {
							throw new UndefiniertException("Es existiert kein Wert für den Zeitpunkt");
						}

						final double mqWert = zuflussWert.doubleValue()
								+ ((qbWert.doubleValue() - anfangsWert.doubleValue())
										* (1.0 - ((double) (zeitPunkt - startZeit) / daempfungsZeit)));
						wert = wert - mqWert;
					}
				}
			}
		}

		// Wert darf nicht kleiner als 0 sein.
		if (wert < 0) {
			wert = 0.0;
		}

		SvpZuflussMenge.LOGGER.finer("Zuflussmenge für MQ: " + mq + " um "
				+ DateFormat.getDateTimeInstance().format(new Date(zeitPunkt)) + " = " + wert);
		return wert;
	}

	/**
	 * ermittelt, ob der übergebene Messquerschnitt innerhalb des Staus der
	 * Zuflussmengenbestimmung liegt.
	 *
	 * @param mqObj
	 *            das MQ-Objekt
	 * @return true, wenn der MQ im Stau liegt
	 */
	private boolean isMqImStau(final SobMessQuerschnitt mqObj) {
		boolean result = false;

		if (mqObj.getSituation() == SobStoerfallSituation.STAU) {
			result = true;
		} else {
			for (final Stau aktuellerStau : mqObj.getRichtung().getStaus().getStauMenge()) {

				final double stauBeginn = aktuellerStau.getStrassenStart();
				final double stauEnde = aktuellerStau.getStrassenEnde();
				final double position = mqObj.getPosition();

				if ((position > stauBeginn) && (position < stauEnde)) {
					result = true;
					break;
				}
			}
		}

		return result;
	}
}
